from functools import cache
import math

import torch


def l1(x): return torch.sum(torch.abs(x))


def sum_squares(x): return torch.sum(x ** 2)


def l2(x):
    # A bit slower but more accurate than torch.linalg.vector_norm
    return torch.sqrt(sum_squares(x))


@cache
def sqrt(n):
    # cached version for things like vectors' dimensions
    return math.sqrt(n)



from cmprs_core.common import *
from cmprs_core.transform import Transform, FOut


class DriveQuantization(Transform):
    """
    This is EDEN specialized to 1-bit
    It should follow a randomized rotation (e.f., Hadamard transform)

    See 'DRIVE: One-bit Distributed Mean Estimation' (NeurIPS 2021) paper:
    https://proceedings.neurips.cc/paper/2021/hash/0397758f8990c1b41b81b43ac389ab9f-Abstract.html
    """

    def forward(self, x) -> FOut:
        # this is the unbiased version of DRIVE
        # an alternative version uses scale = l1(x) / d
        # for minimizing a single vector's MSE

        return FOut(torch.ge(x, 0, out=torch.empty_like(x)), sum_squares(x) / l1(x), bitrate=1)

    def backward(self, assignments, scale):
        return (assignments * 2 - 1) * scale


drive = DriveQuantization()



from typing import Optional

import torch

from cmprs_core.random_utils import dither_like, hook_shared_prng
from cmprs_core.transform import Transform, FunctionalTransform, FOut


class StochasticRounding(Transform):
    def __init__(self, subtractive=False, seed: Optional[int] = None):
        self.subtractive = subtractive

        self.prng = hook_shared_prng(self, seed)

    def forward(self, x) -> FOut:
        """
            :param x: a *normalized* tensor in the [0, 2**bits-1] range
            :return: a tensor with the same dtype in {0, ... , 2**bits - 1},
                     and a random seed for the dither noise if 'subtractive' is True
        """

        return FOut(torch.round(x + dither_like(x, self.prng)))

    def backward(self, tx, _):
        if self.subtractive:
            tx = tx - dither_like(tx, self.prng)

        return tx


round_nearest_transform = FunctionalTransform(lambda x: FOut(torch.round(x)))



import torch

from cmprs_core.transform import Transform, FOut
from cmprs_core.bitrate.recursive_elias_coding import RecursiveEliasCodingBitrate
from cmprs_core.bitrate.optimal import setBitrateFromMeasuredEntropy
from cmprs_core.bitrate.huffman_coding import huffmanCodingBitrate


class SetBitrate(Transform):
    def __init__(self, bitrate: float):
        self.bitrate = bitrate

    def forward(self, x):
        return FOut(tx=x, bitrate=self.bitrate)


class SetBitrateFromTensor(Transform):
    def forward(self, x):
        if x.dtype.is_floating_point:
            bitrate = torch.finfo(x.dtype).bits
        else:
            bitrate = torch.iinfo(x.dtype).bits
        return FOut(tx=x, bitrate=bitrate)


set_bitrate_from_tensor = SetBitrateFromTensor()


class VerifyIntBitrate(Transform):
    def __init__(self, bits):
        self.bits = bits

    def forward(self, x):
        x_int = x.to(torch.int64)
        int_vals = torch.max(x_int - torch.min(x_int)) + 1
        bitrate = torch.ceil(torch.log2(int_vals)).to(torch.int64)
        if self.bits < bitrate:
            raise RuntimeError(f'Input requires {bitrate} bits, but only {self.bits} are allowed '
                               f'(min={torch.min(x)}, max={torch.max(x)}, total value range={int_vals})')
        return FOut(tx=x, bitrate=self.bits)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Sequence, Any, Optional


@dataclass
class FOut:
    """ The return type of the Transform forward method """
    tx: Any
    tcontext: Optional[Any] = None
    bitrate: Optional[float] = None


@dataclass
class ROut:
    rx: Any
    bitrate: Optional[float]


class Transform(ABC):

    @abstractmethod
    def forward(self, x) -> FOut:
        """
        :param x: An object to transform
        :return: FOut Transformed x and additional context, will be used as an input for '_backward'
                      the optional 'bitrate' propagates as a metric
        """

    # noinspection PyMethodMayBeStatic
    def backward(self, tx, tcontext):
        """
        Inverse or a similar operation designed to act on outputs of the stage
        Defaults to no-op
        :param tx: A transformed value returned from the `forward` function
                    parts of it may have passed through other transforms
        :param tcontext: additional context needed for the backward step
        :return: An approximation of the original transformed object
        """
        return tx

    def roundtrip(self, x) -> ROut:
        out = self.forward(x)
        return ROut(self.backward(out.tx, out.tcontext), out.bitrate)


class FunctionalTransform(Transform):
    def __init__(self, forward, backward=None):
        super().__init__()
        self._forward = forward
        if backward is None:
            self._backward = lambda tx, tcontext: tx
        else:
            self._backward = backward

    def forward(self, x) -> FOut:
        return self._forward(x)

    def backward(self, tx, context):
        return self._backward(tx, context)


class SequenceTransform(FunctionalTransform):
    def __init__(self,
                 transforms: Sequence[Transform]):
        """
        :param transforms: these will run in sequence during forward
                           while their backward functions are executed in reverse order during backward

                           we assume that each forward returns a tuple (x, context)
        """

        def _forward(x) -> FOut:
            context = []
            bitrate = None
            for t in transforms:
                f_out = t.forward(x)
                context.append(f_out.tcontext)
                if f_out.bitrate is not None:
                    # in a sequence we trust the latest non-None bitrate
                    bitrate = f_out.bitrate
                x = f_out.tx
            return FOut(x, context, bitrate)

        def _backward(tx, tcontext):
            for t, c in reversed(list(zip(transforms, tcontext))):
                tx = t.backward(tx, c)
            return tx

        super().__init__(_forward, _backward)


class RepeatTransform(SequenceTransform):
    def __init__(self, n: int, transform: Transform):
        super().__init__([transform] * n)


class NoopTransform(Transform):
    def forward(self, x):
        return FOut(x)


noop = NoopTransform()


from typing import Optional

import torch

from cmprs_core.bitrate import SetBitrate, set_bitrate_from_tensor
from cmprs_core.random_utils import bernoulli_mask, hook_shared_prng
from cmprs_core.transform import Transform, FOut


def _mask_split(x, mask):
    x0 = torch.masked_select(x, torch.logical_not(mask))
    x1 = torch.masked_select(x, mask)
    return x0, x1


def _mask_combine(x0, x1, mask):
    x = torch.empty(mask.shape, dtype=x0.dtype, layout=x0.layout, device=x0.device)
    x.masked_scatter_(mask, x1)
    x.masked_scatter_(torch.logical_not(mask), x0)
    return x


class RandomPSplit(Transform):
    def __init__(self, p=0.5, seed: Optional[int] = None,
                 t0: Transform = set_bitrate_from_tensor,
                 t1: Transform = set_bitrate_from_tensor):
        self.p = p
        self.t0 = t0
        self.t1 = t1

        self.prng = hook_shared_prng(self, seed)

    def forward(self, x) -> FOut:
        original_shape = x.shape

        mask = bernoulli_mask(self.p, original_shape, x.device, self.prng)

        x0, x1 = _mask_split(x, mask)
        o0, o1 = self.t0.forward(x0), self.t1.forward(x1)

        if o0.bitrate is None or o1.bitrate is None:
            bitrate = None
        else:
            bitrate = self.p * o0.bitrate + (1 - self.p) * o1.bitrate

        return FOut(
            tx=(o0.tx, o1.tx),
            tcontext=(original_shape, (o0.tcontext, o1.tcontext)),
            bitrate=bitrate,
        )

    def backward(self, tx_s, context):
        tx0, tx1 = tx_s
        original_shape, (c0, c1) = context
        x0, x1 = self.t0.backward(tx0, c0), self.t1.backward(tx1, c1)

        mask = bernoulli_mask(self.p, original_shape, x0.device, self.prng)

        return _mask_combine(x0, x1, mask)


class SignAbsSplit(Transform):
    def __init__(self, t_abs: Transform = set_bitrate_from_tensor, t_sign: Transform = SetBitrate(1)):
        self.t_abs = t_abs
        self.t_sign = t_sign

    def forward(self, x) -> FOut:
        onebit_sign_x = torch.ge(x, 0, out=torch.empty_like(x))
        o_abs, o_sign = self.t_abs.forward(x.abs()), self.t_sign.forward(onebit_sign_x)

        if o_abs.bitrate is None or o_sign.bitrate is None:
            bitrate = None
        else:
            bitrate = o_abs.bitrate + o_sign.bitrate

        return FOut(
            tx=(o_abs.tx, o_sign.tx),
            tcontext=(o_abs.tcontext, o_sign.tcontext),
            bitrate=bitrate,
        )

    def backward(self, tx_s, context):
        abs_tx, sign_tx = tx_s
        abs_c, sign_c = context

        abs_x, onebit_sign_x = self.t_abs.backward(abs_tx, abs_c), self.t_sign.backward(sign_tx, sign_c)
        sign_x = onebit_sign_x * 2 - 1
        return abs_x * sign_x



from abc import ABC, abstractmethod
from typing import Optional

import torch

from cmprs_core.bitrate import set_bitrate_from_tensor
from cmprs_core.random_utils import hook_shared_prng, bernoulli_mask, random_k_mask
from cmprs_core.transform import Transform, FOut


class Sparsification(Transform, ABC):
    def __init__(self, seed: Optional[int] = None, t: Transform = set_bitrate_from_tensor):
        self.t = t
        self.prng = hook_shared_prng(self, seed)

    @abstractmethod
    def random_mask(self, shape, device):
        pass

    @abstractmethod
    def scale(self, x):
        pass

    def forward(self, x):
        original_shape = x.shape

        mask = self.random_mask(x.shape, x.device)
        scale = self.scale(x)

        tx = torch.masked_select(x, mask) * scale
        to = self.t.forward(tx)

        return FOut(
            tx=to.tx,
            tcontext=(original_shape, to.tcontext),
            bitrate=None if to.bitrate is None else to.bitrate / scale
        )

    def backward(self, sparse_tx, tcontext):
        original_shape, tc = tcontext
        sparse_x = self.t.backward(sparse_tx, tc)
        mask = self.random_mask(original_shape, sparse_x.device)

        x = torch.zeros(original_shape, dtype=sparse_x.dtype, device=sparse_x.device).masked_scatter_(mask, sparse_x)
        return x


class RandomP(Sparsification):
    """
    Random Sparsification given a parameter p remaining to determine the
    probability of a coordinate to keep
    """

    def __init__(self, p=0.5, seed: Optional[int] = None, t: Transform = set_bitrate_from_tensor):
        super().__init__(seed, t)
        self.p = p
        self._scale = 1 / self.p

    def random_mask(self, shape, device):
        return bernoulli_mask(self.p, shape, device, self.prng)

    def scale(self, x):
        return self._scale


class RandomK(Sparsification):
    """
    Random Sparsification given a parameter K to determine the number of
    coordinates to keep
    """

    def __init__(self, k, seed: Optional[int] = None, t: Transform = set_bitrate_from_tensor):
        super().__init__(seed, t)
        self.k = k

    def random_mask(self, shape, device):
        return random_k_mask(self.k, shape, device, self.prng)

    def scale(self, x):
        return x.numel() / self.k



from dataclasses import dataclass, asdict
from typing import Any, Optional

import numpy as np
import torch

import cmprs_core.config as config
from cmprs_core.transform import Transform, FOut


@dataclass
class PRNG:
    torch: Any = None
    np: Any = None
    state: Any = None

    def reset(self):
        torch_seed, np_state = self.state

        self.np.bit_generator.state = np_state
        self.torch.manual_seed(torch_seed)
        return self



def bernoulli_mask(p: float, shape, device, prng: PRNG):
    """ returns random booleans """
    return torch.empty(shape, dtype=torch.bool, device=device).bernoulli_(p=p, generator=prng.torch)


def random_k_mask(k: int, shape, device, prng: PRNG):
    """ returns a random boolean tensor with exactly k True entries """
    d = np.prod(shape)
    mask = np.zeros(shape, dtype=bool)
    mask.put(prng.np.choice(d, size=k, replace=False), True)
    return torch.from_numpy(mask).to(device)


def dither_like(x, prng: PRNG):
    """ returns random reals in [-0.5, 0.5) """
    # having an open right bound is important for stochastic rounding
    # in order to avoid rounding the dither up when we get 2**bits-1 + 0.5
    return torch.rand(x.shape, dtype=x.dtype, device=x.device, generator=prng.torch) - 0.5


def rademacher_like(x, prng: PRNG):
    """ returns a random vector in {-1, 1}**d """
    return torch.empty_like(x).bernoulli_(generator=prng.torch) * 2 - 1


UINT64_MAX = np.iinfo(np.uint64).max


def full_range_uint64(np_prng):
    return np_prng.integers(0, UINT64_MAX, dtype=np.uint64, endpoint=True).item()


def hook_shared_prng(transform: Transform, seed: Optional[int] = None):
    if seed is not None:
        seed = [config.root_seed, seed]
    ss = np.random.SeedSequence(seed)
    device = config.device

    np_prng = np.random.default_rng(ss)
    torch_prng = torch.Generator(device=device)

    prng = PRNG(torch=torch_prng, np=np_prng)

    t_forward = transform.forward
    t_backward = transform.backward

    def _forward(x):
        torch_seed = full_range_uint64(prng.np)

        np_state = prng.np.bit_generator.state
        prng.torch.manual_seed(torch_seed)

        prng.state = torch_seed, np_state

        return FOut(**asdict(t_forward(x)))

    def _backward(tx, context):
        prng.reset()

        return t_backward(tx, context)

    transform.forward = _forward
    transform.backward = _backward

    return prng



import math
from typing import Literal

from cmprs_core.clipping import ternGradClip
from cmprs_core.reshape import flatten
from cmprs_core.bitrate import VerifyIntBitrate, RecursiveEliasCodingBitrate
from cmprs_core.rotations.hadamard import padToPowerOf2, RandomizedHadamard
from cmprs_core.rounding import StochasticRounding, round_nearest_transform
from cmprs_core.scale import minmax_normalization, ConstScale, UnitNormalization
from cmprs_core.sparsification import RandomP
from cmprs_core.split import RandomPSplit, SignAbsSplit
from cmprs_core.transform import SequenceTransform as Seq


def _with_factional_bits_support(quantization_fn, bits, sparsity_seed=None):
    """

    :param quantization_fn: a quantization that only support natural numbers, expects one parameter: natural bits
    :param bits: a float representing possible fractional bit rates
    :param sparsity_seed:
    :return:
    """
    if bits < 1:
        # sparsify to support sub 1 bit rates
        return RandomP(bits, sparsity_seed, t=Seq([quantization_fn(1), VerifyIntBitrate(1)]))
    else:
        bits_frac, bits_low = math.modf(bits)
        bits_low = int(bits_low)
        if bits_frac <= 1e-5:  # it's not worth the effort if frac less than 1 in 100_000
            # use standard quantization
            return quantization_fn(bits_low)
        else:
            # random split to support fractional bit rates
            # the fraction determines how many will be assigned to bits_low
            return RandomPSplit(bits_frac, sparsity_seed,
                                t0=Seq([quantization_fn(bits_low), VerifyIntBitrate(bits_low)]),
                                t1=Seq([quantization_fn(bits_low + 1), VerifyIntBitrate(bits_low + 1)]))


RoundingType = Literal['randomized', 'subtractive', 'nearest']


def rounding(rounding_type: RoundingType, bits, normalization=minmax_normalization, rounding_seed=None,
             sparsity_seed=None, rotation=False, rotation_seed=None):
    if rounding_type == 'randomized':
        _rounding = StochasticRounding(subtractive=False, seed=rounding_seed)
    elif rounding_type == 'subtractive':
        _rounding = StochasticRounding(subtractive=True, seed=rounding_seed)
    elif rounding_type == 'nearest':
        _rounding = round_nearest_transform
    else:
        raise ValueError(f'Invalid {rounding_type=}')

    def _quantization_fn(b):
        return Seq([normalization, ConstScale(2 ** b - 1), _rounding])

    r = _with_factional_bits_support(_quantization_fn, bits, sparsity_seed)

    if rotation:
        r = Seq([flatten, padToPowerOf2, RandomizedHadamard(rotation_seed), r])

    return r


def qsgd(bits, norm_ord=2, rounding_seed=None, sparsity_seed=None):
    if bits <= 1 + 1e-5:
        raise ValueError('QSGD require more than 1 bit')

    r = Seq([rounding(rounding_type='randomized',
                      bits=bits - 1,  # one bit will be used to the sign
                      normalization=UnitNormalization(norm_ord=norm_ord),
                      rounding_seed=rounding_seed,
                      sparsity_seed=sparsity_seed), RecursiveEliasCodingBitrate(bits - 1)])

    return SignAbsSplit(t_abs=r)


def terngrad(rounding_seed=None):
    return Seq([
        ternGradClip,
        UnitNormalization(norm_ord=float('inf')),
        StochasticRounding(subtractive=False, seed=rounding_seed),
        VerifyIntBitrate(2)
    ])



import math
from typing import Literal, Type, Optional

import torch

from cmprs_core.bitrate import setBitrateFromMeasuredEntropy, SetBitrate, VerifyIntBitrate
from cmprs_core.common import l2, sqrt, sum_squares
from cmprs_eden.drive import drive
from cmprs_eden.quantization.entropy_delta import EntropyDeltaQuantizer
from cmprs_eden.quantization.opt import OptQuantizer
from cmprs_eden.quantization.quantizer import Quantizer
from cmprs_core.random_utils import hook_shared_prng, bernoulli_mask
from cmprs_core.reshape import flatten
from cmprs_core.rotations.hadamard import RandomizedHadamard, padToPowerOf2
from cmprs_core.sparsification import RandomP
from cmprs_core.split import RandomPSplit, _mask_split, _mask_combine
from cmprs_core.transform import Transform, SequenceTransform as Seq, FOut

QuantizationTypes = Literal['opt', 'entropy_delta']


class EdenExpectedScale(Transform):
    def __init__(self, Q: Type[Quantizer], bits_low, bits_frac=None):
        self.ep = Q.data.ep_for_bits(bits_low, bits_frac)

        # EDEN's scale = l2(x)^2 / <R(x), Q(R(x) * sqrt(d) / l2(x))>
        # Since R(x) * sqrt(d) / l2(x) ~ Normal(0, 1); we get in expectation:
        # scale = l2(x)^2 / ((l2(x) / sqrt(d)) * d * E(Z * Q(Z)))
        # = (l2(x)/sqrt(d)) / E(Z * Q(Z)) = std / ep

    def forward(self, x) -> FOut:
        d = x.numel()

        # post-rotation standard deviation:
        std = l2(x) / sqrt(d)

        return FOut(x / std, std / self.ep)

    def backward(self, tx, scale):
        return tx * scale


class EdenQ(Transform):
    def __init__(self, q: Quantizer):
        self.q = q

    def forward(self, x) -> FOut:
        return FOut(self.q.assign_clusters(x), x.dtype)

    def backward(self, assignments, orig_dtype):
        return self.q.to_values(assignments).to(orig_dtype)


class EdenFull(Transform):
    def __init__(self, q: Quantizer):
        self.q = q

    def forward(self, x) -> FOut:
        std = l2(x) / sqrt(x.numel())
        assignments = self.q.assign_clusters(x / std)
        values = self.q.to_values(assignments)
        x64 = x.to(torch.float64)
        scale = sum_squares(x64) / (values @ x64)
        return FOut(assignments, (scale, x.dtype))

    def backward(self, assignments, context):
        scale, orig_dtype = context
        return (self.q.to_values(assignments) * scale).to(orig_dtype)


class EdenOptFullSplit(Transform):
    def __init__(self, bits_low: int, bits_frac: float, seed: Optional[int] = None):
        self.q_s = OptQuantizer(bits_low), OptQuantizer(bits_low + 1)
        self.p = bits_frac

        self.prng = hook_shared_prng(self, seed)

    def forward(self, x) -> FOut:
        std = l2(x) / sqrt(x.numel())
        original_shape = x.shape

        mask = bernoulli_mask(self.p, original_shape, x.device, self.prng)
        assignments = [q.assign_clusters(_x) for q, _x in zip(self.q_s, _mask_split(x / std, mask))]
        values = _mask_combine(*[q.to_values(_a) for q, _a in zip(self.q_s, assignments)], mask)

        x64 = x.to(torch.float64)
        scale = sum_squares(x64) / (values @ x64)
        return FOut(assignments, (scale, x.dtype, original_shape))

    def backward(self, assignments, context):
        scale, orig_dtype, original_shape = context
        mask = bernoulli_mask(self.p, original_shape, assignments.device, self.prng)
        values = _mask_combine(*[q.to_values(_a) for q, _a in zip(self.q_s, assignments)], mask)
        return (values * scale).to(orig_dtype)


def eden(bits: int, quantization: QuantizationTypes = 'opt', use_expected_scale=True, split_seed: int = None,
         rotation_seed: int = None):
    rotation = [flatten, padToPowerOf2, RandomizedHadamard(rotation_seed)]

    if quantization == 'entropy_delta':
        if use_expected_scale:
            e = [EdenExpectedScale(EntropyDeltaQuantizer, bits), EdenQ(EntropyDeltaQuantizer(bits))]
        else:
            e = [EdenFull(EntropyDeltaQuantizer(bits))]
        return Seq([*rotation, *e, setBitrateFromMeasuredEntropy])

    if quantization != 'opt':
        raise ValueError(f'Unexpected {quantization=}')

    #  quantization == 'opt'

    if bits <= 1 + 1e-5:
        if use_expected_scale:
            e = [EdenExpectedScale(OptQuantizer, 1), EdenQ(OptQuantizer(1)), VerifyIntBitrate(1)]
        else:
            e = [drive]  # same as EdenFull(OptQuantizer(1))
        rest = Seq([*rotation, *e])
        if bits < 1 - 1e-5:
            # sparsify
            return RandomP(bits, split_seed, t=rest)
        else:
            # one-bit
            return rest
    else:
        # More than 1 bit
        bits_frac, bits_low = math.modf(bits)
        bits_low = int(bits_low)
        if bits_frac <= 1e-5:
            # integer bits
            if use_expected_scale:
                e = [EdenExpectedScale(OptQuantizer, bits_low), EdenQ(OptQuantizer(bits_low))]
            else:
                e = [EdenFull(OptQuantizer(bits_low))]
            return Seq([*rotation, *e, VerifyIntBitrate(bits_low)])
        else:
            # factional bits
            if use_expected_scale:
                return Seq([
                    *rotation,
                    EdenExpectedScale(OptQuantizer, bits_low, bits_frac),
                    RandomPSplit(bits_frac, split_seed,
                                 t0=Seq([EdenQ(OptQuantizer(bits_low)), VerifyIntBitrate(bits_low)]),
                                 t1=Seq([EdenQ(OptQuantizer(bits_low + 1)), VerifyIntBitrate(bits_low + 1)])),
                ])
            else:
                return Seq([*rotation, EdenOptFullSplit(bits_low, bits_frac, split_seed), SetBitrate(bits)])



from cmprs_core.transform import Transform, FOut


class FlattenTransform(Transform):
    def forward(self, x):
        return FOut(x.view(-1), x.shape)

    def backward(self, tx, original_shape):
        return tx.view(original_shape)


flatten = FlattenTransform()



import torch

from cmprs_core.transform import Transform, FOut


class SetBitrateFromMeasuredEntropy(Transform):

    def forward(self, x):
        if x.is_floating_point():
            x_int = x.to(torch.int64)
        else:
            x_int = x
        x_int = x_int - torch.min(x_int)  # start at 0, avoid negative values
        freq = torch.bincount(x_int)
        freq = freq[freq.nonzero(as_tuple=True)]
        probs = freq / x.numel()

        # This is the entropy, which is the optimal bitrate:
        # https://en.wikipedia.org/wiki/Shannon%27s_source_coding_theorem
        # there are coders such as arithmetic coding that come very close
        # (https://en.wikipedia.org/wiki/Arithmetic_coding)
        bitrate = torch.sum(-probs * torch.log2(probs)).item()

        return FOut(tx=x, bitrate=bitrate)


setBitrateFromMeasuredEntropy = SetBitrateFromMeasuredEntropy()


from collections.abc import Sequence
from dataclasses import dataclass
from functools import cache
from typing import List

import torch
from scipy.stats import norm

import cmprs_core.config as config
from cmprs_core.transform import Transform, FOut


def calc_expected_round_product(has_zero_cluster, delta, stop_at_prob=0):
    # for X ~ N(0, (1 / delta) ** 2)
    # when has_zero_cluster=True --> calculate E[X*Round(X)]
    # when has_zero_cluster=False --> calculate E[X * (Round(X + ½ sign(X)) - ½ sign(X))]
    p = float('inf')
    product = 0
    if has_zero_cluster:
        n = 1 / 2
        # we will safely skip the 0..0.5 interval since it rounds to zero
    else:
        n = 0
    while p > stop_at_prob:
        l, r = n, (n + 1)
        e = norm.expect(lb=l, ub=r,
                        scale=1 / delta)  # this is the same as centroid * probability_mass (since centroid == e/probability_mass)
        product += e * (n + 1 / 2)
        n += 1
        p = norm.sf(r, scale=1 / delta)  # survival, which is more numerically stable 1 - cdf
    return 2 * product  # multiply by 2 since we only covered positive values


def calc_expected_round_product2(has_zero_cluster, delta, stop_at_prob=0):
    product = 0
    if has_zero_cluster:
        n = 1 / 2
        # we can safely skip the 0..0.5 where centroid and round center agree
    else:
        n = 0

    l, r = n, (n + 1)
    p = (norm.cdf(r, scale=1 / delta) - norm.cdf(l, scale=1 / delta))
    while p > stop_at_prob:
        centroid = norm.expect(lb=l, ub=r,
                               scale=1 / delta) / p  # this is the same as centroid * probability_mass (since centroid === e/probability_mass)
        product += p * centroid ** 2
        n += 1
        l, r = n, (n + 1)
        p = (norm.cdf(r, scale=1 / delta) - norm.cdf(l, scale=1 / delta))
    return 2 * product


def calc_expected_normal_bias_correction(has_zero_cluster, delta, stop_at_prob=0):
    if has_zero_cluster:
        n = 1 / 2
        # we can safely skip the 0..0.5 where centroid and round center agree
    else:
        n = 0
    centroid_mad = round_center_mad = 0

    l, r = n, (n + 1)
    p = (norm.cdf(r, scale=1 / delta) - norm.cdf(l, scale=1 / delta))
    while p > stop_at_prob:
        centroid_mad += norm.expect(lb=l, ub=r,
                                    scale=1 / delta)  # this is the same as centroid * probability_mass (since centroid === e/probability_mass)
        round_center_mad += p * (n + 1 / 2)
        n += 1
        l, r = n, (n + 1)
        p = (norm.cdf(r, scale=1 / delta) - norm.cdf(l, scale=1 / delta))
    return centroid_mad / round_center_mad


def calc_expected_deviation_ratio(has_zero_interval, delta, stop_at_interval_mass=1e-15, num_of_positive_intervals=float('inf')):
    intervals = pos_normal_intervals(has_zero_interval, delta, stop_at_interval_mass, num_of_positive_intervals)[0]

    c_deviation, lc_deviation = zip(*[(interval.mass * interval.c, interval.mass * interval.lc) for interval in intervals])

    return sum(c_deviation) / sum(lc_deviation)


def calc_expected_products(has_zero_interval, delta, stop_at_interval_mass=1e-15, num_of_positive_intervals=float('inf')):
    intervals = pos_normal_intervals(has_zero_interval, delta, stop_at_interval_mass, num_of_positive_intervals)[0]

    c_product, lc_product = zip(*[(interval.mass * interval.c ** 2, interval.mass * interval.c * interval.lc) for interval in intervals])

    return sum(c_product), sum(lc_product)


@dataclass
class Interval:
    lb: float  # lower bound
    ub: float  # upper bound
    c: float  # centroid (conditional expectation at (lb, ub))
    lc: float  # linear center (lb + delta / 2)
    mass: float  # cdf(ub) - cdf(lb)


@dataclass
class Intervals:
    intervals: List[Interval]
    clip_prob: float  # probability of falling outside [-(lb+delta), (lb+delta)] for max lower bound
    # for X ~ N(0, (1 / delta) ** 2)
    expected_lc_product: float  # E[X*Round(X)] or E[X * (Round(X + ½ sign(X)) - ½ sign(X))] depending on has_zero_interval
    expected_c_product: float   # E[X*Q(X)] where Q is the centroid of the interval containing X:
                                #    for has_zero_interval = True
                                #           (-2.5, -1.5), (-1.5, -0.5), (-0.5, 0.5), (0.5, 1.5), (1.5, 2.5), ...
                                #           or
                                #    for has_zero_interval = False
                                #           (Round(X + ½ sign(X)) - ½ sign(X) - ½, Round(X + ½ sign(X)) - ½ sign(X) + ½)

@cache
def pos_normal_intervals(has_zero_interval, delta, stop_at_interval_mass=1e-15, num_of_positive_intervals=float('inf')):
    intervals: List[Interval] = []
    scale = 1 / delta
    if has_zero_interval:
        n = 1 / 2
    else:
        n = 0

    last_interval = False
    while not last_interval:
        lb, ub = n, n + 1
        ub_sf = norm.sf(ub, scale=scale)
        if n + 1 >= num_of_positive_intervals or ub_sf <= stop_at_interval_mass:
            # there will be no right interval point
            # effectively we clip values at upper bound
            clip_prob = 2 * ub_sf
            ub = float('inf')
            ub_sf = 0
            last_interval = True
        mass = norm.sf(lb, scale=scale) - ub_sf

        c = norm.expect(lb=lb, ub=ub, scale=scale, conditional=True)
        lc = lb + 1 / 2
        intervals.append(Interval(lb, ub, c, lc, mass))

        n += 1

    return intervals, clip_prob


def calc_centroids(has_zero_interval, delta,
                   stop_at_interval_mass=0, num_of_positive_intervals=float('inf')):
    max_c = 0
    centroids = []
    scale = 1 / delta
    if has_zero_interval:
        n = 1 / 2
    else:
        n = 0

    lb = n
    if n + 1 >= num_of_positive_intervals:
        # there is no right interval point
        ub = float('inf')
    else:
        ub = n + 1
    cdf_lb, cdf_ub = norm.cdf(lb, scale=scale), norm.cdf(ub, scale=scale)
    interval_mass = cdf_ub - cdf_lb

    while interval_mass > stop_at_interval_mass:
        centroid = norm.expect(lb=lb, ub=ub, scale=scale, conditional=True)
        if centroid <= max_c:
            # numeric instability...
            break
        else:
            max_c = centroid
        centroids.append(delta * centroid)

        n += 1

        lb = n
        if n + 1 >= num_of_positive_intervals:
            # there is no right interval point
            ub = float('inf')
        else:
            ub = n + 1
        cdf_lb, cdf_ub = cdf_ub, norm.cdf(ub, scale=scale)
        interval_mass = cdf_ub - cdf_lb

    negative_centroids = [-c for c in reversed(centroids)]
    if has_zero_interval:
        return [*negative_centroids, 0, *centroids]
    else:
        return [*negative_centroids, *centroids]


@cache
def calc_correction(has_zero_cluster, delta, stop_at_prob=0):
    # We divide by the expected product and multiply
    # by the constant normal bias correction that exists
    # from rounding to equal intervals instead of centroids
    # Also note that we divide by delta and not delta**2
    # since multiplying by delta is needed for final rescaling
    a = calc_expected_normal_bias_correction(has_zero_cluster, delta, stop_at_prob)
    b = calc_expected_round_product2(has_zero_cluster, delta, stop_at_prob)
    return a / (b * delta)


def round_to_half_int(x):
    """round the input to the nearest ½𝓩"""
    half_sign = torch.sign(x) / 2
    # round towards nearest ½𝓩
    # even integers go towards zero, uneven integers go away from zero
    x = torch.round(x + half_sign) - half_sign
    # we randomly round zeros towards {-0.5, 0.5}
    zeros_mask = x.eq(0)
    n_zeros = torch.count_nonzero(zeros_mask)
    dither = torch.round(torch.rand(n_zeros)) - 0.5  # random values in {-0.5, 0.5}
    return x.masked_scatter_(zeros_mask, dither)


class LinearEden(Transform):
    def __init__(self, sqrt_2_raw_moment, has_zero_cluster, delta, stop_at_prob=0):
        self.has_zero_cluster = has_zero_cluster
        self.delta = delta

        # EDEN's scale = l2(x)^2 / <R(x), Q(R(x) * sqrt(d) / l2(x))>
        #   Q(R(x) * sqrt(d) / l2(x)) -> Round(R(x) * sqrt(d) / l2(x) / delta) * delta
        #   Since (R(x) * sqrt(d) / l2(x) / delta) ~ N(0, (1 / delta) ** 2)
        #   we have <R(x), Q(R(x) * sqrt(d) / l2(x))> = (l2(x) / sqrt(d)) * delta * d * E[X * Q(X)] * delta
        #   => l2(x)^2 / <R(x), Q(R(x) * sqrt(d) / l2(x)) = (l2(x)/sqrt(d))/ (delta**2 * E[X * Q(X)])
        #                                                 = std / (delta**2 * ep)

        # for linearity we also need an estimate of input coordinate std
        # since post-Hadamard std is l2/sqrt(d) and E(l2) = sqrt(d * E(x**2)) for coordinate x
        # the expected post-Hadamard std is sqrt(E(x**2)) (this is sqrt of the 2-raw-moment)
        # (if the input vector coordinates have zero mean this is their std)
        self.std = sqrt_2_raw_moment  # should be the pre-hadamard coordinates' sqrt(E(x**2))

        # this is std * normal_corr / (ep * delta)
        self.constant_scale = self.std * calc_correction(has_zero_cluster, delta, stop_at_prob)

    def forward(self, x) -> FOut:
        tx = x / self.std / self.delta
        if self.has_zero_cluster:
            tx = torch.round(tx)
        else:
            tx = round_to_half_int(tx)
        # (instead of multiplying tx by delta here, the constant scale is divided by delta)

        return FOut(tx, self.constant_scale)

    def backward(self, tx, scale):
        return tx * scale


class DeltaCentroidQuantizer:
    def __init__(self, has_zero_cluster, delta, stop_at_prob=0):
        centroids = calc_centroids(has_zero_cluster, delta, stop_at_prob)
        self.centroids = torch.tensor(centroids, dtype=torch.float64).to(config.device)
        self.n_values = len(self.centroids)

        self.has_zero_cluster = has_zero_cluster
        self.delta = delta

    def assign_clusters(self, x):
        """
        :param x: vector to be clustered
        :return: The cluster assignment vector (torch.int64).
                 The output includes negative values, and it is assumed
                 that a (near) optimal entropy coder follow.
        """

        tx = x / self.delta
        if self.has_zero_cluster:
            tx = torch.round(tx)
        else:
            tx = round_to_half_int(tx)

        x = (tx + self.n_values // 2 + (-0.5 if not self.has_zero_cluster else 0)).to(torch.int64)
        return torch.clamp(x, min=0, max=self.n_values - 1)

    def to_values(self, assignments):
        """
        :param assignments: cluster indices as returned by assign_clusters
        :return: each cluster index replaced with the cluster's centroid (torch.float64)
        """
        # theoretically, since the normal distribution is infinite we can get an assignment
        # value for which we did not calculate the centroid... in practice this does not happen
        # since we calculated many very un-probable centroids
        return torch.take(self.centroids, assignments)



from functools import cache
from math import floor, log2

import torch

from cmprs_core.transform import Transform, FOut


# recursive_elias a.k.a.elias omega coding

@cache
def recursive_elias_code_len(n):
    if n < 1:
        raise ValueError('elias omega code only accept positive integers')
    if n == 1:
        return 1
    num_bits = floor(log2(n)) + 1
    return num_bits + recursive_elias_code_len(num_bits - 1)


@cache
def elias_n_2_code_len(bits, device):
    return torch.tensor([recursive_elias_code_len(n) for n in range(1, 2 ** bits + 1)], dtype=torch.float32,
                        device=device)


def elias_bitrate(x, bits):
    #  assumes x is in [0, 2**bits - 1]
    n_2_code_len = elias_n_2_code_len(bits, x.device)
    if x.is_floating_point():
        x = torch.round(x)
    return torch.mean(torch.take(n_2_code_len, x.to(torch.int64)))


class RecursiveEliasCodingBitrate(Transform):
    def __init__(self, bits):
        self.bits = bits

    def forward(self, x) -> FOut:
        bitrate = elias_bitrate(x, self.bits)

        return FOut(tx=x, bitrate=bitrate)



import torch

from cmprs_core.transform import Transform, FOut


class TernGradClip(Transform):
    def forward(self, x):
        th = 2.5 * torch.std(x)
        clipped_vec = torch.clamp(x, -th, th)

        return FOut(clipped_vec)


ternGradClip = TernGradClip()



import torch

from cmprs_core.transform import Transform, FOut


class MinMaxNormalization(Transform):
    """
    Maps a real vector into [0..1] range: (x - min(x)) / (max(x) - min(x))
    """

    def forward(self, x) -> FOut:
        minimum = x.min()
        maximum = x.max()

        return FOut(torch.nan_to_num((x - minimum) / (maximum - minimum)), (minimum, maximum))

    def backward(self, tx, context):
        minimum, maximum = context
        return tx * (maximum - minimum) + minimum


minmax_normalization = MinMaxNormalization()


class UnitNormalization(Transform):
    """
    Normalizes a real vector by dividing it by its norm (the resulting vector's norm is 1)
    """

    def __init__(self, norm_ord):
        self.ord = norm_ord

    def forward(self, x) -> FOut:
        x_norm = torch.linalg.vector_norm(x, ord=self.ord)

        return FOut(torch.nan_to_num(x / x_norm), x_norm)

    def backward(self, tx, x_norm):
        return tx * x_norm


l1_normalization = UnitNormalization(norm_ord=1)
l2_normalization = UnitNormalization(norm_ord=2)
max_normalization = UnitNormalization(norm_ord=float('inf'))


class ConstScale(Transform):
    def __init__(self, scale):
        self.scale = scale

    def forward(self, x) -> FOut:
        return FOut(x * self.scale)

    def backward(self, tx, _):
        return tx / self.scale


class ConstOffset(Transform):
    def __init__(self, offset):
        self.offset = offset

    def forward(self, x) -> FOut:
        return FOut(x + self.offset)

    def backward(self, tx, _):
        return tx - self.offset



from heapq import heappush, heappop, heapify

import torch

from cmprs_core.transform import Transform, FOut


def huffman_bitrate(probs):
    # min heap - the lowest combined probability at the top
    # each heap node contains a list of per-integer probabilities
    heap = [(p, [p]) for p in probs if p > 0]

    # build the heap
    heapify(heap)

    # iteratively combine two heap nodes with the lowest frequencies
    bitrate = 0
    while len(heap) > 1:
        key_1, probs_1 = heappop(heap)
        key_2, probs_2 = heappop(heap)

        # We add together the integer probabilities each time we would otherwise add a bit
        # to their representing string
        bitrate += sum(probs_1)
        bitrate += sum(probs_2)

        heappush(heap, (key_1 + key_2, probs_1 + probs_2))

    return bitrate


class HuffmanCodingBitrate(Transform):

    def forward(self, x) -> FOut:
        if x.is_floating_point():
            x_int = x.to(torch.int64)
        else:
            x_int = x
        probs = (torch.bincount(x_int) / x.numel()).cpu().numpy()

        bitrate = huffman_bitrate(probs)

        return FOut(tx=x, bitrate=bitrate)


huffmanCodingBitrate = HuffmanCodingBitrate()



from typing import Optional

import numpy as np
import torch

from cmprs_core.random_utils import hook_shared_prng
from cmprs_core.rotations.hadamard import randomized_hadamard_transform, inverse_randomized_hadamard_transform
from cmprs_core.transform import Transform, FOut


def _kashin_padded_dim(dim, pad_threshold):
    if not dim & (dim - 1) == 0:
        padded_dim = int(2 ** (np.ceil(np.log2(dim))))
        if dim / padded_dim > pad_threshold:
            padded_dim = 2 * padded_dim
    else:
        padded_dim = 2 * dim

    return padded_dim


class KashinRepresentation(Transform):

    def __init__(self, eta=0.9, delta=1.0, pad_threshold=0.85, n_iters=3, seed: Optional[int] = None):
        self.prng = hook_shared_prng(self, seed)

        self.eta = eta
        self.delta = delta
        self.pad_threshold = pad_threshold
        self.n_iters = n_iters

    def forward(self, x):
        dim = x.numel()
        padded_dim = _kashin_padded_dim(dim, self.pad_threshold)

        kashin_coefficients = torch.zeros(padded_dim, device=x.device)
        padded_x = torch.zeros(padded_dim, device=x.device)

        M = torch.norm(x) / np.sqrt(self.delta * padded_dim)

        for i in range(self.n_iters):
            padded_x[:] = 0
            padded_x[:dim] = x
            padded_x = randomized_hadamard_transform(padded_x, self.prng.reset())

            b = padded_x
            b_hat = torch.clamp(b, min=-M, max=M)

            kashin_coefficients = kashin_coefficients + b_hat

            if i < self.n_iters - 1:
                b_hat = inverse_randomized_hadamard_transform(b_hat, self.prng.reset())
                x = x - b_hat[:dim]

                M = self.eta * M

        return FOut(kashin_coefficients, dim)

    def backward(self, qhvec, dim):
        return inverse_randomized_hadamard_transform(qhvec, self.prng)[:dim]
